from jsonrpc._json import dumps
from django.http import HttpResponse
from django.shortcuts import render_to_response
from jsonrpc.site import jsonrpc_site
from jsonrpc import mochikit


def browse(request):
    if request.GET.get('f', None) == 'mochikit.js':
        return HttpResponse(mochikit.mochikit,
                            content_type='application/x-javascript')
    if request.GET.get('f', None) == 'interpreter.js':
        return HttpResponse(mochikit.interpreter,
                            content_type='application/x-javascript')
    desc = jsonrpc_site.service_desc()
    return render_to_response('browse.html', {
        'methods': desc['procs'],
        'method_names_str': dumps(
            [m['name'] for m in desc['procs']])
    })



import sys
from uuid import uuid1
from jsonrpc._json import loads, dumps
from jsonrpc.exceptions import *
from jsonrpc._types import *
from django.conf import settings
from django.core import signals
from django.utils.encoding import smart_text
from django.core.serializers.json import DjangoJSONEncoder


def empty_dec(f):
    """
    An empty decorator
    :param f: function
    :return: function
    """
    return f


try:
    from django.views.decorators.csrf import csrf_exempt
except (NameError, ImportError):
    csrf_exempt = empty_dec


encode_kw = lambda to_encode: {str(key): value for key, value in to_encode.items()}


def trim_docstring(docstring):
    if not docstring:
        return ''
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)
    # Return a single string:
    return '\n'.join(trimmed)


def encode_kw11(to_encode):
    """
    The function to encode kwargs.
    :param: to_encode: kwargs to encode.
    :return: encoded kwargs: dict
    """
    encoded_kw = {}
    if not isinstance(to_encode, dict):
        return encoded_kw
    
    for k, v in to_encode.items():
        try:
            int(k)
        except ValueError:
            encoded_kw.update({k: v})
    return encoded_kw

def encode_arg11(to_encode):
    """
    The function to encode args.
    :param: to_encode: args to encode.
    :return: encoded args: dict
    """
    encoded_args = set()
    if isinstance(to_encode, list):
        return to_encode
    elif isinstance(to_encode, dict):
        _ = encode_kw(to_encode)
        for key in _.keys():
            try:
                int(key)
            except ValueError:
                pass
            else:
                encoded_args.add(str(key))
        
        encoded_args = [_[key] for key in sorted(encoded_args)]
        return encoded_args

def validate_params(method, attributes):
    """
    Function to validate parameters.
    :param method: Type of HTTP method.
    :param attributes: parameters.
    """
    print(method.json_arg_types.keys())
    if type(attributes['params']) == Object:
        keys = method.json_arg_types.keys()
        if len(keys) != len(attributes['params']):
            raise InvalidParamsError('Not enough params provided for '
                                     '{}'.format(
                                         method.json_sig))
        for key in keys:
            if not key in attributes['params']:
                print('\n\n\n\nSHITTER SHITTER', key, attributes, '\n\n\n\n')
                raise InvalidParamsError("{} is not a valid parameter for "
                                         "{}".format(key, method.json_sig))
            if not Any.kind(
                    attributes['params'][key])==method.json_arg_types[key]:
                raise InvalidParamsError('{} is not the correct type {} '
                                         'for {}'.format(
                                             type(attributes['params'][key]),
                                             method.json_arg_types[key],
                                             method.json_sig))
    elif type(attributes['params']) == Array:
        print(attributes)
        arg_types = list(method.json_arg_types.values())
        try:
            for index, arg in enumerate(attributes['params']):
                if not Any.kind(arg) == arg_types[index]:
                    raise InvalidParamsError('{} is not the correct type '
                                             '{} for {}'.format(
                                                 type(arg), arg_types[index],
                                                 method.json_sig))
        except IndexError:
            raise InvalidParamsError('Too many params provided for '
                                     '{}'.format(method.json_sig))
        else:
            if len(attributes['params']) != len(arg_types):
                raise InvalidParamsError('Not enough params provided '
                                         'for {}'.format(method.json_sig))


class JsonRpcSite(object):
    """A JSON-RPC Site"""

    def __init__(self, json_encoder=DjangoJSONEncoder):
        """
        Constructor for JsonRpcSite.
        :param json_encoder: type of Json Encoder. Default to DjangoJSONEncoder
        """
        self.urls = {}
        self.uuid = uuid1()
        self.version = '1.0'
        self.name = 'django-json-rpc'
        self.register('system.describe', self.describe)
        self.json_encoder = json_encoder
        self.available_versions = ['2.0', '1.1', '1.0']

    def _empty_response(self, version='1.0'):
        """
        Method to provide empty response
        :param version: Version of JSON RPC
        :return: Empty Response
        """
        response = {'id': None}
        if version == '2.0':
            response.update({'error': None, 'result': None, 'jsonrpc': version})
        if version == '1.1':
            response['version'] = version
        return response

    def _validate_get(self, request, method):
        """
        :param request: Request instance
        :param method: http method
        :return: tuple of "if request is valid and data.
        """
        if not request.method == "GET":
            return False, {}

        encoded_params = lambda req: {key: value[0] if len(value) == 1 else value for key, value in req}
        method = smart_text(method)
        if method in self.urls and getattr(self.urls[method], 'json_safe', False):
            data = {
                'params': encoded_params(request.GET.lists()),
                'method': method,
                'id': 'jsonrpc',
                'version': '1.1'
            }
            return True, data

    def _response_dict(self, request, data, is_batch=False, version='1.0',
                       json_encoder=None):
        """
        Method to get the response for the request.
        :param request: request for json rpc
        :param data: parameters associated with the request
        :param is_batch: defaults to false
        :param version: JSON RPC version
        :param json_encoder: Json Encoder
        :return: response, status code
        """
        json_encoder = json_encoder or self.json_encoder
        response = self._empty_response(version=version)

        def apply_version(function, req, param, version):
            if version == '1.1':
                return function(req, *encode_arg11(param),
                                **encode_kw(encode_kw11(param)))
            if isinstance(param, dict):
                return function(req, **encode_kw(param))
            return function(req, *param)

        try:
            if 'params' not in data:
                data['params'] = []
            if 'method' not in data:
                raise InvalidParamsError("Request requires str: 'method' and "
                                         "list: 'params'.")
            if data['method'] not in self.urls:
                raise MethodNotFoundError("Method not found, "
                                          "Available methods:\n\t{}".format(
                                              "\n\t".join(self.urls.keys())))

            v_key = ''
            if 'jsonrpc' in data:
                v_key = 'jsonrpc'
            elif 'version' in data:
                v_key = 'version'

            try:

                if str(data[v_key]) not in self.available_versions:
                    raise InvalidRequestError(
                        "JSON-RPC version {} not supported.Please raise bug @ "
                        "https://github.com/Rishi-jha/django-json-rpc".format(
                            data[v_key]))
                version = request.jsonrpc_version = response[v_key] = str(data['version'])
            except KeyError:
                request.jsonrpc_version = '1.0'

            method = self.urls[str(data['method'])]
            if getattr(method, 'json_validate', False):
                validate_params(method, data)

            if 'id' in data and data['id'] is not None:
                response['id'] = data['id']
                if version in ('1.1', '2.0') and 'error' in response:
                    response.pop('error')
            elif is_batch:
                raise InvalidRequestError("Not ok in batch format")

            _result = apply_version(method, request, data['params'], version)
            if not data.get('id'):
                return None, 204
            _result = list(_result) if isinstance(_result, tuple) else _result
            encoder = json_encoder()
            _builtin_types = set((dict, list, set, NoneType, bool, six.text_type) + six.integer_types + six.string_types)
            if not any(isinstance(_result, ty) for ty in _builtin_types):
                try:
                    encoder.default(_result)
                except KeyError:
                    raise TypeError("Return type not supported for {}".format(_result))

            response['result'] = _result
            status = 200

        except Error as e:
            response['error'] = e.json_rpc_format
            if version in ('1.1', '2.0') and 'result' in response:
                response.pop('result')
            status = e.status

        except Exception as e:

            signals.got_request_exception.send(sender=self.__class__,
                                               request=request)

            # Put stacktrace into the OtherError only if DEBUG is enabled
            if settings.DEBUG:
                other_error = OtherError(e)
            else:
                other_error = OtherError("Internal Server Error")

            response['error'] = other_error.json_rpc_format
            status = other_error.status
            if version in ('1.1', '2.0') and 'result' in response:
                response.pop('result')

        # Exactly one of result or error MUST be specified. It's not
        # allowed to specify both or none.
        if version in ('1.1', '2.0') and not response.get('error'):
            response.pop('error')

        return response, status

    @csrf_exempt
    def dispatch(self, request, method='', json_encoder=None):
        """
        dispatch method for registering apps
        :param request: Request to process
        :param method: type of method
        :param json_encoder: json encdoer
        :return: response
        """
        from django.http import HttpResponse
        json_encoder = json_encoder or self.json_encoder


        try:
            response = self._empty_response()
            if request.method.upper() == 'GET':
                is_valid, data = self._validate_get(request, method)
                if not is_valid:
                    raise InvalidRequestError("The method you are trying to "
                                              "access is not available by GET "
                                              "requests")
            elif not request.method.upper() == 'POST':
                raise RequestPostError
            else:
                try:
                    if hasattr(request, "body"):
                        data = loads(request.body.decode('utf-8'))
                    else:
                        data = loads(request.raw_post_data.decode('utf-8'))
                except:
                    raise InvalidRequestError
            if isinstance(data, list):
                response = [self._response_dict(request, data=d,
                                                is_batch=True,
                                                json_encoder=json_encoder)[0]
                            for d in data]
                status = 200
            else:
                response, status = self._response_dict(
                    request, data,
                    json_encoder=json_encoder)
                if response is None and not data.get('id'):
                    return HttpResponse('', status=status)

            json_rpc = dumps(response, cls=json_encoder)

        except Error as e:
            response['error'] = e.json_rpc_format
            status = e.status
            json_rpc = dumps(response, cls=json_encoder)
        except Exception as e:
            # exception missed by others
            signals.got_request_exception.send(sender=self.__class__,
                                               request=request)

            # Put stacktrace into the OtherError only if DEBUG is enabled
            if settings.DEBUG:
                other_error = OtherError(e)
            else:
                other_error = OtherError("Internal Server Error")

            response['result'] = None
            response['error'] = other_error.json_rpc_format
            status = other_error.status

            json_rpc = dumps(response, cls=json_encoder)

        return HttpResponse(json_rpc,
                            status=status,
                            content_type='application/json-rpc')


    def procedure_desc(self, key):
        """
        describing the procedure
        :param key: method name
        :return: dict
        """
        method = self.urls[key]
        return {
            'name': method.json_method,
            'summary': trim_docstring(method.__doc__),
            'idempotent': method.json_safe,
            'params': [{'type': str(Any.kind(t)),
                        'name': k} for k, t in method.json_arg_types.items()],
            'return': {'type': str(method.json_return_type)}
        }

    def service_desc(self):
        """
        Describing the service
        :return: dict
        """
        return {
            'sdversion': '1.0',
            'name': self.name,
            'id': 'urn:uuid:{}'.format(self.uuid),
            'summary': trim_docstring(self.__doc__),
            'version': self.version,
            'procs': [self.procedure_desc(key) for key in self.urls.keys()
                      if self.urls[key] != self.describe]
        }

    def describe(self, request):
        return self.service_desc()

    def register(self, name, method):
        self.urls[smart_text(name)] = method



jsonrpc_site = JsonRpcSite()



import uuid
from six.moves.urllib import request as urllib_request
from six.moves.urllib import error as urllib_error
from django.test.client import FakePayload

from ._json import loads, dumps
from ._types import *


class ServiceProxy(object):
    def __init__(self, service_url, service_name=None, version='1.0'):
        self.version = str(version)
        self.service_url = service_url
        self.service_name = service_name

    def __getattr__(self, name):
        if self.service_name != None:
            name = "{}.{}".format(self.service_name, name)
        params = dict(self.__dict__, service_name=name)
        return self.__class__(**params)

    def __str__(self):
        return "ServiceProxy {}".format({
            'jsonrpc': self.version,
            'method': self.service_name
        })

    def send_payload(self, params):
        """Performs the actual sending action and returns the result"""
        data = dumps({
            'jsonrpc': self.version,
            'method': self.service_name,
            'params': params,
            'id': str(uuid.uuid1())
        }).encode()
        print(data)
        headers = {
            'Content-Type': 'application/json-rpc',
            'Accept': 'application/json-rpc',
            'Content-Length': len(data)
        }
        req = urllib_request.Request(self.service_url, data, headers)
        print(req)
        resp = urllib_request.urlopen(req)
        print(resp)
        try:
            req = urllib_request.Request(self.service_url, data, headers)
            resp = urllib_request.urlopen(req)
        except IOError as e:
            print(e)
            if isinstance(e, urllib_error.HTTPError):
                if e.code not in (
                        401, 403
                ) and e.headers['Content-Type'] == 'application/json-rpc':
                    return e.read().decode('utf-8')  # we got a jsonrpc-formatted respnose
                raise ServiceProxyException(e.code, e.headers, req)
            else:
                raise e
        return resp.read().decode('utf-8')

    def __call__(self, *args, **kwargs):
        params = kwargs if len(kwargs) else args
        if Any.kind(params) == Object and self.version != '2.0':
            raise Exception('Unsupported arg type for JSON-RPC 1.0 '
                            '(the default version for this client, '
                            'pass version="2.0" to use keyword arguments)')

        r = self.send_payload(params)
        y = loads(r)
        if 'error' in y:
            try:
                from django.conf import settings
                if settings.DEBUG:
                    print('JSONRPC: {} error {}'.format(self.service_name, y))
            except:
                pass
        return y


class ServiceProxyException(IOError):
    def __init__(self, code, headers, request):
        self.args = ('An Error Occurred', code, headers, request)
        self.code = code
        self.message = 'An Error Occurred'
        self.headers = headers
        self.request = request


class TestingServiceProxy(ServiceProxy):
    """Service proxy which works inside Django unittests"""

    def __init__(self, client, *args, **kwargs):
        super(TestingServiceProxy, self).__init__(*args, **kwargs)
        self.client = client

    def send_payload(self, params):
        json_data = dumps({
            'jsonrpc': self.version,
            'method': self.service_name,
            'params': params,
            'id': str(uuid.uuid1())
        })
        json_payload = FakePayload(json_data)
        client_args = {
            'wsgi.input': json_payload,
            'CONTENT_LENGTH': len(json_data)
        }
        response = self.client.post(self.service_url, **client_args)
        return response.content.decode('utf-8')



try:
    from django.utils.translation import ugettext_lazy as _
    _("You're lazy...")  # this function lazy-loads settings
except (ImportError, NameError):
    _ = lambda t, *a, **k: t
from django.utils.encoding import smart_text


class Error(Exception):
    """ Error class based on the JSON-RPC 2.0 specs
      http://groups.google.com/group/json-rpc/web/json-rpc-1-2-proposal

      code    - number
      message - string
      data    - object

      status  - number    from http://groups.google.com/group/json-rpc/web/json-rpc-over-http JSON-RPC over HTTP Errors section
  """

    code = 0
    message = None
    data = None
    status = 500

    def __init__(self, message=None):
        """ Setup the Exception and overwrite the default message """
        if message is not None:
            self.message = message

    @property
    def json_rpc_format(self):
        """ return the Exception data in a format for JSON-RPC """

        error = {
            'name': smart_text(self.__class__.__name__),
            'code': self.code,
            'message': "{}: {}".format(smart_text(self.__class__.__name__),
                                       smart_text(self.message)),
            'data': self.data
        }

        from django.conf import settings

        if settings.DEBUG:
            import sys, traceback
            error['stack'] = traceback.format_exc()
            error['executable'] = sys.executable

        return error

# Exceptions
# from http://groups.google.com/group/json-rpc/web/json-rpc-1-2-proposal

# The error-codes -32768 .. -32000 (inclusive) are reserved for pre-defined errors.
# Any error-code within this range not defined explicitly below is reserved for future use


class ParseError(Error):
    """ Invalid JSON. An error occurred on the server while parsing the JSON text. """
    code = -32700
    message = _('Parse error.')


class InvalidRequestError(Error):
    """ The received JSON is not a valid JSON-RPC Request. """
    code = -32600
    message = _('Invalid Request.')
    status = 400


class MethodNotFoundError(Error):
    """ The requested remote-procedure does not exist / is not available. """
    code = -32601
    message = _('Method not found.')
    status = 404


class InvalidParamsError(Error):
    """ Invalid method parameters. """
    code = -32602
    message = _('Invalid params.')


class ServerError(Error):
    """ Internal JSON-RPC error. """
    code = -32603
    message = _('Internal error.')

# -32099..-32000    Server error.     Reserved for implementation-defined server-errors.

# The remainder of the space is available for application defined errors.


class RequestPostError(InvalidRequestError):
    """ JSON-RPC requests must be POST """
    message = _('JSON-RPC requests must be POST')


class InvalidCredentialsError(Error):
    """ Invalid login credentials """
    code = 401
    message = _('Invalid login credentials')
    status = 401


class OtherError(Error):
    """ catchall error """
    code = 500
    message = _('Error missed by other exceptions')
    status = 500



import threading

from django.contrib.auth.models import User
from django.core import management

from jsonrpc._json import loads, dumps
from six.moves.urllib import request as urllib_request
from six.moves.urllib import parse as urllib_parse

from jsonrpc.proxy import ServiceProxy

TEST_DEFAULTS = {
    'ROOT_URLCONF': 'test.jsontesturls',
    'DEBUG': True,
    'DEBUG_PROPAGATE_EXCEPTIONS': True,
    'DATETIME_FORMAT': 'N j, Y, P',
    'USE_I18N': False,
    'INSTALLED_APPS': (
        'jsonrpc',
        'django.contrib.auth',
        'django.contrib.contenttypes',
        'django.contrib.sessions'),
    'DATABASES': {
        'default': {
            'ENGINE': 'django.db.backends.sqlite3',
            'NAME': 'test.sqlite3',
        },
    },
    'MIDDLEWARE_CLASSES': (
        'django.middleware.common.CommonMiddleware',
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django.middleware.csrf.CsrfViewMiddleware',
        'django.contrib.auth.middleware.AuthenticationMiddleware',
    ),
    'AUTHENTICATION_BACKENDS': ('django.contrib.auth.backends.ModelBackend',),
    'TEMPLATE_LOADERS': (
        'django.template.loaders.filesystem.load_template_source',
        'django.template.loaders.app_directories.load_template_source'),
}

from django.conf import settings
settings.configure(**TEST_DEFAULTS)

import django
if hasattr(django, 'setup'):
    # Run django.setup() for Django>=1.7
    django.setup()

def _call(host, req):
    return loads(urllib_request.urlopen(host, dumps(req).encode('utf-8')).read().decode('utf-8'))


class JsonServer(object):
    def _thread_body(self):
        try:
            from wsgiref.simple_server import make_server
            from django.core.handlers.wsgi import WSGIHandler
            management.call_command('migrate', interactive=False)
            try:
                User.objects.create_user(username='rishi', email='rishijha424@gmail.com', password='password').save()
            except Exception:
                pass

            http = make_server('', 8999, WSGIHandler())
            print('Server created. continue={0}'.format(self.continue_serving))
            self.event.set() # notify parent thread that the server is ready to serve requests
            while self.continue_serving:
                print('Waiting for request!')
                http.handle_request()
                self.n_requests += 1
                print('Handled {0} requests!'.format(self.n_requests))
            print('Got server stop! requests={0}'.format(self.n_requests))
            http.server_close()
            print('Server closed!')
        except Exception as e:
            import traceback
            traceback.print_exc()
            print('Error starting server: {0}'.format(e))
        finally:
            if not self.event.is_set():
                self.event.set()

    def start(self):
        print('Starting Server...')
        self.continue_serving = True
        self.n_requests = 0
        self.event = threading.Event()
        self.t = threading.Thread(target=self._thread_body)
        self.t.start()
        self.event.wait()
        return self

    def stop(self):
        print('Stopping Server...')
        self.continue_serving = False
        try:
            proxy = ServiceProxy('http://127.0.0.1:8999/json/', version=2.0)
            proxy.jsonrpc.test(string='Hello')['result']
        except: # doesnt matter if this fails
            pass
        self.t.join(2.0)
        return self




#!/usr/bin/env python

import os

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

HERE = os.path.dirname(os.path.abspath(__file__))

setup(
    name="json-rpc-django",
    version="1.0.1",
    description=("A simple JSON-RPC implementation for Django. An extension of " 
                 "django-json-rpc by Samuel Sutch for python 3"),
    author="Rishikesh Jha",
    author_email="rishijha424@gmail.com",
    license="MIT",
    url="https://github.com/Rishi-jha/django-json-rpc",
    download_url="https://github.com/Rishi-jha/json_rpc_django/archive/1.0.1.tar.gz",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment', 'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent', 'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Framework :: Django', 'Framework :: Django :: 1.4',
        'Framework :: Django :: 1.5', 'Framework :: Django :: 1.6',
        'Framework :: Django :: 1.7', 'Framework :: Django :: 1.8',
        'Framework :: Django :: 2.1',
    ],
    packages=['jsonrpc'],
    zip_safe=False,  # we include templates and tests
    install_requires=['Django>=2.0', 'six'],
    package_data={'jsonrpc': ['templates/*']})



import re
import sys
from inspect import getargspec
from jsonrpc.site import jsonrpc_site
from jsonrpc._types import *
from jsonrpc.exceptions import *

from collections import OrderedDict


default_site = jsonrpc_site
KWARG_RE = re.compile(
    r'\s*(?P<arg_name>[a-zA-Z0-9_]+)\s*=\s*(?P<arg_type>[a-zA-Z]+)\s*$')
SIG_RE = re.compile(
    r'\s*(?P<method_name>[a-zA-Z0-9._]+)\s*(\((?P<args_sig>[^)].*)?\)'
    r'\s*(\->\s*(?P<return_sig>.*))?)?\s*$')


class JSONRPCTypeCheckingUnavailable(Exception):
    pass


def _type_checking_available(sig='', validate=False):
    if not hasattr(type, '__eq__') and validate:  # and False:
        raise JSONRPCTypeCheckingUnavailable(
            'Type checking is not available in your version of Python '
            'which is only available in Python 2.6 or later. Use Python 2.6 '
            'or later or disable type checking in {}'.format(sig))


def _validate_arg(value, expected):
    """Returns whether or not ``value`` is the ``expected`` type."""
    return isinstance(value, expected)


def _eval_arg_type(arg_type, T=Any, arg=None, sig=None):
    """
    Returns a type from a snippet of python source. Should normally be
    something just like 'str' or 'Object'.

      arg_type      the source to be evaluated
      T             the default type
      arg           context of where this type was extracted
      sig           context from where the arg was extracted

    Returns a type or a Type
    """
    try:
        T = eval(arg_type)
    except Exception as e:
        raise ValueError(
            'The type of {} could not be evaluated in {} for {}: {}'.format(
                arg_type, arg, sig, str(e)))
    else:
        if type(T) not in (type, Type):
            raise TypeError(
                '{} is not a valid type in {} for {}'.format(
                    repr(T), arg, sig))
        return T


def _parse_sig(sig, arg_names, validate=False):
    """
    Parses signatures into a ``OrderedDict`` of paramName => type.
    Numerically-indexed arguments that do not correspond to an argument
    name in python (ie: it takes a variable number of arguments) will be
    keyed as the stringified version of it's index.

      sig         the signature to be parsed
      arg_names   a list of argument names extracted from python source

    Returns a tuple of (method name, types dict, return type)
    """
    d = SIG_RE.match(sig)
    if not d:
        raise ValueError('Invalid method signature {}'.format(sig))
    d = d.groupdict()
    ret = [(n, Any) for n in arg_names]
    if ('args_sig' in d and isinstance(d['args_sig'], str)
            and d['args_sig'].strip()):
        for i, arg in enumerate(d['args_sig'].strip().split(',')):
            _type_checking_available(sig, validate)
            if '=' in arg:
                if not isinstance(ret, OrderedDict):
                    ret = OrderedDict(ret)
                dk = KWARG_RE.match(arg)
                if not dk:
                    raise ValueError('Could not parse arg type {} '
                                     'in {}'.format(arg, sig))
                dk = dk.groupdict()
                if not sum(
                    [(k in dk and type(dk[k]) is str and bool(dk[k].strip()))
                         for k in ('arg_name', 'arg_type')]):
                    raise ValueError('Invalid kwarg value {} in {}'.format(
                        arg, sig))
                ret[dk['arg_name']] = _eval_arg_type(dk['arg_type'], None, arg,
                                                     sig)
            else:
                if type(ret) is OrderedDict:
                    raise ValueError('Positional arguments must occur '
                                     'before keyword arguments in {}'.format(sig))
                if len(ret) < i + 1:
                    ret.append((str(i), _eval_arg_type(arg, None, arg, sig)))
                else:
                    ret[i] = (ret[i][0], _eval_arg_type(arg, None, arg, sig))
    if not type(ret) is OrderedDict:
        ret = OrderedDict(ret)
    return (d['method_name'], ret,
            (_eval_arg_type(d['return_sig'], Any, 'return', sig) if
             d['return_sig'] else Any))


def _inject_args(sig, types):
    """
    A function to inject arguments manually into a method signature before
    it's been parsed. If using keyword arguments use 'kw=type' instead in
    the types array.

      sig     the string signature
      types   a list of types to be inserted

    Returns the altered signature.
    """
    if '(' in sig:
        parts = sig.split('(')
        sig = '{}({}{}{}'.format(
            parts[0], ', '.join(types),
            (', ' if parts[1].index(')') > 0 else ''), parts[1])
    else:
        sig = '{}({})'.format(sig, ', '.join(types))
    return sig


def jsonrpc_method(name,
                   authenticated=False,
                   authentication_arguments=['username', 'password'],
                   safe=False,
                   validate=False,
                   site=default_site):
    """
    Wraps a function turns it into a json-rpc method. Adds several attributes
    to the function specific to the JSON-RPC machinery and adds it to the default
    jsonrpc_site if one isn't provided. You must import the module containing
    these functions in your urls.py.

        name

            The name of your method. IE: `namespace.methodName` The method name
            can include type information, like `ns.method(String, Array) -> Nil`.

        authenticated=False

            Adds `username` and `password` arguments to the beginning of your
            method if the user hasn't already been authenticated. These will
            be used to authenticate the user against `django.contrib.authenticate`
            If you use HTTP auth or other authentication middleware, `username`
            and `password` will not be added, and this method will only check
            against `request.user.is_authenticated`.

            You may pass a callable to replace `django.contrib.auth.authenticate`
            as the authentication method. It must return either a User or `None`
            and take the keyword arguments `username` and `password`.

        safe=False

            Designates whether or not your method may be accessed by HTTP GET.
            By default this is turned off.

        validate=False

            Validates the arguments passed to your method based on type
            information provided in the signature. Supply type information by
            including types in your method declaration. Like so:

            @jsonrpc_method('myapp.specialSauce(Array, String)', validate=True)
            def special_sauce(self, ingredients, instructions):
              return SpecialSauce(ingredients, instructions)

            Calls to `myapp.specialSauce` will now check each arguments type
            before calling `special_sauce`, throwing an `InvalidParamsError`
            when it encounters a discrepancy. This can significantly reduce the
            amount of code required to write JSON-RPC services.

        site=default_site

            Defines which site the jsonrpc method will be added to. Can be any
            object that provides a `register(name, func)` method.

    """

    def decorator(func):
        arg_names = getargspec(func)[0][1:]
        X = {'name': name, 'arg_names': arg_names}
        if authenticated:
            if authenticated is True or six.callable(authenticated):
                # TODO: this is an assumption
                X['arg_names'] = authentication_arguments + X['arg_names']
                X['name'] = _inject_args(X['name'], ('String', 'String'))
                from django.contrib.auth import authenticate as _authenticate
                from django.contrib.auth.models import User
            else:
                authenticate = authenticated

            @six.wraps(func)
            def _func(request, *args, **kwargs):
                user = getattr(request, 'user', None)
                is_authenticated = getattr(user, 'is_authenticated',
                                           lambda: False)
                if ((user is not None and six.callable(is_authenticated) and
                     not is_authenticated()) or user is None):
                    user = None
                    try:
                        creds = args[:len(authentication_arguments)]
                        if len(creds) == 0:
                            raise IndexError
                        # Django's authenticate() method takes arguments as dict
                        user = _authenticate(username=creds[0],
                                             password=creds[1], *creds[2:])
                        if user is not None:
                            args = args[len(authentication_arguments):]
                    except IndexError:
                        auth_kwargs = {}
                        try:
                            for auth_kwarg in authentication_arguments:
                                auth_kwargs[auth_kwarg] = kwargs[auth_kwarg]
                        except KeyError:
                            raise InvalidParamsError(
                                'Authenticated methods require at least '
                                '[{arguments}] or {{{arguments}}} '
                                'arguments'.format(
                                    arguments=', '.join(
                                        authentication_arguments)))

                        user = _authenticate(**auth_kwargs)
                        if user is not None:
                            for auth_kwarg in authentication_arguments:
                                kwargs.pop(auth_kwarg)
                    if user is None:
                        raise InvalidCredentialsError
                    request.user = user
                return func(request, *args, **kwargs)
        else:
            _func = func

        @six.wraps(_func)
        def exc_printer(*a, **kw):
            try:
                return _func(*a, **kw)
            except Exception as e:
                try:
                    print('JSONRPC SERVICE EXCEPTION')
                    import traceback
                    traceback.print_exc()
                except:
                    pass
                six.reraise(*sys.exc_info())

        ret_func = exc_printer

        method, arg_types, return_type = \
        _parse_sig(X['name'], X['arg_names'], validate)
        ret_func.json_args = X['arg_names']
        ret_func.json_arg_types = arg_types
        ret_func.json_return_type = return_type
        ret_func.json_method = method
        ret_func.json_safe = safe
        ret_func.json_sig = X['name']
        ret_func.json_validate = validate
        site.register(method, ret_func)
        return ret_func

    return decorator



import six
from functools import reduce


def _types_gen(T):
    yield T
    if hasattr(T, 't'):
        for l in T.t:
            yield l
            if hasattr(l, 't'):
                for ll in _types_gen(l):
                    yield ll


def _basetypes(T):
    return [t for t in _types_gen(T) if not isinstance(t, Type)]


class Type(type):
    """ A rudimentary extension to `type` that provides polymorphic
  types for run-time type checking of JSON data types. IE:

  assert type(u'') == String
  assert type('') == String
  assert type('') == Any
  assert Any.kind('') == String
  assert Any.decode('str') == String
  assert Any.kind({}) == Object
  """

    def __init__(self, *args, **kwargs):
        type.__init__(self, *args, **kwargs)

    def __eq__(self, other):
        if six.PY2:
            for T in _types_gen(self):
                if isinstance(other, Type):
                    if T in other.t:
                        return True
                if type.__eq__(T, other):
                    return True
        elif six.PY3:
            for T in _types_gen(self):
                if isinstance(other, Type) and T in other.t:
                    return True
                elif other in _basetypes(T):
                    return True
        return False

    def __str__(self):
        return "{}({})".format(getattr(self, '_name', 'unknown'), getattr(
            self, 't', None))

    def N(self, n):
        self._name = n
        return self

    def I(self, *args):
        self.t = list(args)
        return self

    def kind(self, t):
        if type(t) is Type:
            return t
        ty = lambda t: type(t)
        if type(t) is type:
            ty = lambda t: t
        return reduce(
            lambda L, R: R if (hasattr(R, 't') and ty(t) == R) else L,
            [T for T in _types_gen(self) if T is not Any])

    def decode(self, n):
        return reduce(lambda L, R: R if (str(R) == n) else L, _types_gen(self))


str_types = (six.text_type, )
if six.PY2:
    str_types += (six.binary_type, )
num_types = six.integer_types + (float, )

# JSON primitives and data types
Object = Type('Object', (object, ), {}).I(dict).N('obj')
Number = Type('Number', (object, ), {}).I(*num_types).N('num')
Boolean = Type('Boolean', (object, ), {}).I(bool).N('bit')
String = Type('String', (object, ), {}).I(*str_types).N('str')
Array = Type('Array', (object, ), {}).I(list, set, tuple).N('arr')
Nil = Type('Nil', (object, ), {}).I(type(None)).N('nil')
Any = Type('Any', (object, ), {}).I(
    Object, Number, Boolean, String, Array, Nil).N('any')
NoneType = type(None)



import json

try:
    json.dumps
    json.loads
except AttributeError:
    try:  # monkey patching for python-json package
        json.dumps = lambda obj, *args, **kwargs: json.write(obj)
        json.loads = lambda str, *args, **kwargs: json.read(str)
    except AttributeError:
        raise ImportError('Could not load an appropriate JSON library '
                          'currently supported are simplejson, '
                          'python3.7 json and python-json')

loads = json.loads
dumps = json.dumps
